library(dplyr)
library(statmod)

# STEP 1 - assumptions
# STEP 1.1 - assumptions applying to all simulations
# number observations
n_max<-150000   
# poisson parameter
lambda_avg=.02
# gamma parameters
gamma_scale=50000
gamma_shape=2
# prevalence protected class 
pct_protected = 0.333
# random holdout size
pct_holdoutdata=0.5
# p parameter for tweedies
p_parm=1.3
# combine into a single dataframe
sim_parms<-cbind(n_max,lambda_avg,gamma_scale,gamma_shape,pct_protected,pct_holdoutdata)%>%
  as.data.frame()%>%mutate_all(as.character)%>%mutate_all(as.numeric)
rm(n_max,lambda_avg,gamma_scale,gamma_shape,pct_protected,pct_holdoutdata)

# STEP 1.2 - assumptions that vary by scenario
# the next four arrays should each have the same number of elements
# the _th element of each array corresponds to scenario _
# scenario "tags" (labels)
phi    <-c('N','P','M','PM','C','PC','MC','PMC')
# predilection constants
delta  <-c(0,  0.5, 0,    0.5,  0,    0.5,  0,    0.5)
# misalignment constants
epsilon<-c(0,  0,   0.5,  0.5,  0,    0,    0.5,  0.5)
# confusion constants
alpha  <-c(0,  0,   0,    0,    0.5,  0.5,  0.5,  0.5)
# combine into a single dataframe
bias_parms<-as.data.frame(cbind(phi,epsilon,delta,alpha))%>%
  mutate_all(as.character)%>%mutate_at(c('epsilon','delta','alpha'),as.numeric)%>%
  bind_cols(sim_parms)
rm(phi,epsilon,delta,alpha,sim_parms)


# PRIMARY LOOP: conducts simulation and modeling steps across the _ scenarios
for (phi_counter in 1:nrow(bias_parms)){
  
  # extract parameters from PART 1 that are pertinent to scenario being run
  delta=bias_parms$delta[phi_counter]
  epsilon=bias_parms$epsilon[phi_counter]
  alpha=bias_parms$alpha[phi_counter]
  n_max=bias_parms$n_max[phi_counter]
  lambda_avg=bias_parms$lambda_avg[phi_counter]
  gamma_scale=bias_parms$gamma_scale[phi_counter]
  gamma_shape=bias_parms$gamma_shape[phi_counter]
  pct_protected=bias_parms$pct_protected[phi_counter]
  pct_holdout=bias_parms$pct_holdout[phi_counter]
  phi=bias_parms$phi[phi_counter]
  
  # STEP 2 - SIMULATION
  # set random seed for reproducibility
  set.seed(123)
  # STEP 2.1 create exposure dataframe of desired size for use in modeling
  exposures<-runif(n_max,0,1)%>%as.data.frame()
  # STEP 2.2 conduct six random draws that are starting
  exposures<-exposures%>%rename(r_phi_0_n=1)
  exposures$r_phi_3_n<-runif(n_max,0,1)  # speeding
  exposures$r_phi_1_n<-runif(n_max,0,1)  # density
  exposures$r_phi_2_n<-runif(n_max,0,1)  # age
  exposures$r_phi_5_n<-runif(n_max,0,1)  # safety
  exposures$r_phi_4_n<-runif(n_max,0,1)  # weight
  exposures$n<-1:nrow(exposures) 
  
  # STEP 2.3 = experiential draw
  exposures<-exposures%>%
    #  STEP 2.3.1 disturb random draws by predilection influence
    mutate(row=1,
           # disturbance applies to protected class
           protected_tf=ifelse(r_phi_0_n<=pct_protected,1,0),
           predilection=1+ifelse(protected_tf==1,delta,0),
           x_phi_3_n=r_phi_3_n/predilection,
           x_phi_1_n=r_phi_1_n/predilection,
           x_phi_5_n=r_phi_5_n/predilection
    )%>%
    # STEP 2.3.2 rank normalize experiential draw
    arrange(x_phi_3_n)%>%mutate(x_phi_3_n=cumsum(row)/n_max)%>%
    arrange(x_phi_1_n)%>%mutate(x_phi_1_n=cumsum(row)/n_max)%>%
    arrange(x_phi_5_n)%>%mutate(x_phi_5_n=cumsum(row)/n_max)%>%
    
    # STEP 2.4 = social draw
    # STEP 2.4.1 disturb experiential draws by misalignment influence
    mutate(misalignment=1+ifelse(protected_tf==1,epsilon,0),
           s_phi_3_n=x_phi_3_n/misalignment,
           s_phi_1_n=x_phi_1_n/misalignment,
           s_phi_5_n=x_phi_5_n/misalignment
    )%>%
    
    # STEP 2.4.2 rank normalize social draw
    arrange(s_phi_3_n)%>%mutate(s_phi_3_n=cumsum(row)/n_max)%>%
    arrange(s_phi_1_n)%>%mutate(s_phi_1_n=cumsum(row)/n_max)%>%
    arrange(s_phi_5_n)%>%mutate(s_phi_5_n=cumsum(row)/n_max)%>%
    
    # STEP 2.5 use social draw to assign rating variable levels
    # cut points of inequalities and factors in 2.5-6 reconcile to paper
    mutate(territory1=ifelse(s_phi_1_n <= 0.333,1,0),  
           territory3=ifelse(s_phi_1_n > 0.667,1,0), 
           drvage1=ifelse(r_phi_2_n <= 0.1,1,0),  
           drvage3=ifelse(r_phi_2_n > 0.8,1,0), 
           speeding1=ifelse(s_phi_3_n <= 0.1,1,0),
           vehicle1=ifelse(r_phi_4_n<=0.25,1,0), 
           vehicle3=ifelse(r_phi_4_n>0.75,1,0), 
           safety1=ifelse(s_phi_5_n<=0.2,1,0),  
           
           # STEP 2.6 use experiential draw to assign rating variable impacts
           # these will serve as poisson/gamma parameters
           F_1_n=ifelse(x_phi_1_n <= 0.333,1.5,
                        ifelse(x_phi_1_n > 0.667, 0.5, 1)),
           F_2_n=ifelse(r_phi_2_n <= 0.1,1.9,
                        ifelse(r_phi_2_n > 0.8,1.6,0.7)),
           F_3_n=ifelse(x_phi_3_n <= 0.1,2,.89),
           F_4_n=ifelse(r_phi_4_n <= 0.25,1.2,ifelse(r_phi_4_n > 0.75,0.8,1)),
           F_5_n=ifelse(x_phi_5_n <= 0.2,1.2,0.95),
           F_0_n=ifelse(protected_tf==1,1+alpha/2,1-alpha/4))
  
  # STEP 2.7 conduct random draw from poisson on each exposure
  # mean used in simulation is based on user input baseline mean and rating variable impacts
  exposures=exposures%>%mutate(lambda=lambda_avg*F_3_n*F_1_n*F_2_n)
  claims_lambda<-exposures%>%ungroup()%>%select(lambda)
  exposures$avg_frq <- mapply(rpois, n = nrow(exposures), lambda = claims_lambda)
  rm(claims_lambda)
  
  # STEP 2.8  conduct random draw from gamma on each claim
  # STEP 2.8.1 scale parameter used in simulation on user input baseline mean and rating variable impacts
  exposures = exposures %>% mutate(theta=gamma_scale*F_5_n*F_4_n*F_0_n)

  # STEP 2.8.2 create claims dataframe consisting of only exposures with non-zero poisson
  claims<-exposures%>%filter(avg_frq>0)%>%mutate(gamma=theta)
  # STEP 2.8.3 use replication to create one row for each claim based on the poisson
  claims <- claims[rep(seq_len(nrow(claims)), times = claims$avg_frq), ]
  claims_theta<-claims%>%ungroup()%>%select(gamma)
  # STEP 2.8.4 random draw
  claims$avg_sev <- mapply(rgamma, n = nrow(claims), shape=gamma_shape, scale = claims_theta)
  # STEP 2.8.5 aggregate losses from claims level sim back to exposure level
  claims<-claims%>%mutate(clm_ct=1)%>%group_by(n)%>%summarize(loss_amt=sum(avg_sev),clm_ct=sum(clm_ct))
  rm(claims_theta)
  # STEP 2.8.6 join losses back to exposures, changing missing to 0
  n2z<-function(x){x<-ifelse(is.na(x),0,x)}
  exposures<-exposures%>%left_join(claims)%>%mutate_at(c('loss_amt','clm_ct'),n2z)
  rm(claims)
  
  # STEP 3 - GLMs
  # set random seed for reproducibility, random draw to tag holdout observations
  set.seed(123)
  exposures$r_phi_6_n<-runif(nrow(exposures),0,1)
  exposures=exposures%>%mutate(cohort=ifelse(r_phi_6_n<=pct_holdout,'training','holdout'))
  
  # STEP 3.0 - converge Sociodemographic GLM using protected class status
  socio_glm<-glm('loss_amt ~ protected_tf ',exposures%>%filter(cohort=='training'), 
                 family='tweedie'(var.power = p_parm, link.power = 0))
  # parse coefficients and error statistics out of model object
  socio_coeff<-coef(summary(socio_glm))%>%as.data.frame()
  socio_coeff$var=rownames(socio_coeff)
  rownames(socio_coeff)<-NULL
  socio_coeff$rating_factor=exp(socio_coeff$Estimate)
  socio_coeff$glm=0; socio_coeff$glmname='socio'
  
  # STEP 3.1 - converge Baseline GLM using rating variables
  baseline_str=paste0('loss_amt ~ territory1 + territory3 + 
                      drvage1 + drvage3 + 
                      speeding1 + safety1 + 
                      vehicle1 + vehicle3')
  baseline_glm<-glm(baseline_str,exposures%>%filter(cohort=='training'), 
                    family='tweedie'(var.power = p_parm, link.power = 0))
  baseline_coeff<-coef(summary(baseline_glm))%>%as.data.frame()
  baseline_coeff$var=rownames(baseline_coeff)
  rownames(baseline_coeff)<-NULL
  baseline_coeff$rating_factor=exp(baseline_coeff$Estimate)
  baseline_coeff$glm=1; baseline_coeff$glmname='baseline'
  
  # STEP 3.2 - converge Control Variable GLM using rating variables and protected
  control_str=paste(baseline_str,' + protected_tf')
  control_glm<-glm(control_str, exposures%>%filter(cohort=='training'), 
                   family='tweedie'(var.power = p_parm, link.power = 0))
  control_coeff<-coef(summary(control_glm))%>%as.data.frame()
  control_coeff$var=rownames(control_coeff)
  rownames(control_coeff)<-NULL
  control_coeff$rating_factor=exp(control_coeff$Estimate)
  control_coeff$glm=2; control_coeff$glmname='control'
  
  # STEP 3.3 - converge Residualization GLM using partialled rating variables
  # STEP 3.3.1 - variable preprocessing
  var2partial<-c('territory1','territory3','drvage1','drvage3','speeding1','safety1','vehicle1','vehicle3')
  # step through list of variables to residualize
  for (ctr in 1:length(var2partial)){
    # linearly regress variable against protected class
    # make prediction then store residual under variable name and .par suffix
    partial_str<-paste('partial_glm<-lm(',var2partial[ctr],'~ protected_tf  ,exposures) 
           exposures$',var2partial[ctr],'.par = exposures$',var2partial[ctr],' - predict(partial_glm)', sep='')
    eval(parse(text=partial_str))
    rm(partial_str,partial_glm)}
  
  # STEP 3.3.2 - produce correlation matrix to evaluate residualization
  # filter to only the rating variables
  r2t=function(x){x=round(x,2)}
    cor_df<-exposures%>%
      select(contains("err")|contains("rva")|contains("eed")|
             contains("hic")|contains("fet")|contains("d_t")|
             ends_with("1")|ends_with("2")|ends_with("3")|ends_with("4"))%>%
      select(!contains("pred")&!contains("_fa"))%>%
      # reorder and round to two decimals primarily for cosmetics
      select(ends_with("v")|starts_with("sc")|
            starts_with("p")|!contains(".")|contains("."))%>%
      cor()%>%as.data.frame()%>%mutate(scenario=phi_counter)%>%mutate_all(r2t)
    cor_df$v<-rownames(cor_df)
    rownames(cor_df)<-NULL
    cor_df<-cor_df%>%select(starts_with("sc")|ends_with("v")|!ends_with("v"))
  
  # STEP 3.3.3 - use residualized variables to develop GLM
  resid_str=gsub('3','3.par',gsub('1','1.par',baseline_str))
  resid_glm<-glm(resid_str, exposures%>%filter(cohort=='training'), 
                 family='tweedie'(var.power = p_parm, link.power = 0))
  resid_coeff<-coef(summary(resid_glm))%>%as.data.frame()
  resid_coeff$var=rownames(resid_coeff)
  rownames(resid_coeff)<-NULL
  resid_coeff$rating_factor=exp(resid_coeff$Estimate)
  resid_coeff$glm=3
  resid_coeff$glmname='resid'
  
  # STEP 3.4 - concatenate parsed coefficients into single dataframe
  glm_coef<-bind_rows(socio_coeff,baseline_coeff,control_coeff,resid_coeff)
  rm(socio_coeff,baseline_coeff,control_coeff,resid_coeff)
  gc()
  
  # STEP 4 - EVALUATE PREDICTIONS
  # STEP 4.1 - develop predictions from GLMs
  # STEP 4.1.0 - sociodemographic GLM
  exposures$pred0 = predict(socio_glm,exposures,type='response')
  # STEP 4.1.1 - baseline GLM
  exposures$pred1 = predict(baseline_glm,exposures,type='response')
  # STEP 4.1.2 for control variable GLM, apply balancing
  # predict using rating variable neutral
  exposures$pred21 = predict(control_glm,exposures%>%mutate(protected_tf=0),type='response')
  # predict using actual value of rating variable
  exposures$pred20 = predict(control_glm,exposures,type='response')
  # balancing factor = 1 / ( pure model avg / neutral model avg ), on training
  exposures_train<-exposures%>%filter(r_phi_6_n<=pct_holdout)
  imbalance=mean(exposures_train$pred20)/mean(exposures_train$pred21)
  exposures$pred2 = exposures$pred21*imbalance
  # STEP 4.1.3 - residualized GLM
  exposures$pred3 = predict(resid_glm,exposures,type='response')
  rm(socio_glm,baseline_glm,control_glm,resid_glm,exposures_train,imbalance)
  
  
  # STEP 5 - gini indices
  exposures_holdout=exposures%>%filter(cohort=='holdout')
  gini<-exposures_holdout%>%
    select(row,pred0,pred1,pred2,pred20,pred3,loss_amt,r_phi_6_n,protected_tf)
  gini=bind_rows(gini,gini%>%mutate(protected_tf=9))
  # 5.1 transpose the dataset so each observation has one row per GLM
  gini<-bind_rows(gini%>%mutate(pred=pred0,glm=0),
                  gini%>%mutate(pred=pred1,glm=1),
                  gini%>%mutate(pred=pred2,glm=2),
                  # test control variable GLM had we not neutralized control
                  gini%>%mutate(pred=pred20,glm=2.1),
                  gini%>%mutate(pred=pred3,glm=3))%>%
    select(-pred0,-pred1,-pred2,-pred3)%>%
    # 5.2 Riemanns for Ideal Ginis if we had losses to predict losses
    group_by(glm,protected_tf)%>%arrange(glm,protected_tf,loss_amt,r_phi_6_n)%>%
    mutate(width=1/sum(row),
           diagheight=cumsum(row)/sum(row),
           idealheight=cumsum(loss_amt)/sum(loss_amt),
           idealarea=(diagheight-idealheight)*width)%>%
    # 5.3 Riemanns for Ginis on models themselves
    arrange(glm,protected_tf,pred,r_phi_6_n)%>%
    mutate(diagheight=cumsum(row)/sum(row),
           giniheight=cumsum(loss_amt)/sum(loss_amt),
           giniarea=(diagheight-giniheight)*width)%>%
    summarize(giniindex=2*sum(giniarea),
              giniideal=2*sum(idealarea))%>%
    mutate(pctideal=round(giniindex/giniideal,3))
  gc()
  
  # STEP 6 - Actual vs Expected
  ave=exposures_holdout%>%group_by(protected_tf)%>%
    summarize_at(c('row','loss_amt','pred0','pred1','pred2','pred20','pred3'),sum)
  ave=bind_rows(ave%>%mutate(glm=0,pred=pred0),
                ave%>%mutate(glm=1,pred=pred1),
                ave%>%mutate(glm=2,pred=pred2),
                # test control glm had we not neutralized control
                ave%>%mutate(glm=2.1,pred=pred20),
                ave%>%mutate(glm=3,pred=pred3))
  ave=bind_rows(ave,ave%>%mutate(protected_tf=9))%>%
    group_by(glm,protected_tf)%>%
    summarize_at(c('row','loss_amt','pred'),sum)%>%
    mutate(loss_cost_obs=round(loss_amt/row,2),
           loss_cost_pred=round(pred/row,2),
           relative_error=loss_cost_obs/loss_cost_pred-1.00)%>%
    ungroup()%>%
    mutate(r_9=ifelse(protected_tf==9,row,0),
           r_1=ifelse(protected_tf==1,row,0),  
           r_0=ifelse(protected_tf==0,row,0),
           o_9=ifelse(protected_tf==9,loss_amt,0),
           o_1=ifelse(protected_tf==1,loss_amt,0),  
           o_0=ifelse(protected_tf==0,loss_amt,0),
           p_9=ifelse(protected_tf==9, pred,0),
           p_1=ifelse(protected_tf==1, pred,0),  
           p_0=ifelse(protected_tf==0, pred,0))%>%
    # calculate for cohort relativities to protected, dominant, and total
    group_by(glm)%>%
    mutate_at(c('r_9','r_1','r_0','o_9','o_1','o_0','p_9','p_1','p_0'),sum)%>%
    mutate(obs2dom=loss_cost_obs/(o_0/r_0),
           obs2pro=loss_cost_obs/(o_1/r_1),
           obs2tot=loss_cost_obs/(o_9/r_9),
           pred2dom=loss_cost_pred/(p_0/r_0),
           pred2pro=loss_cost_pred/(p_1/r_1),
           pred2tot=loss_cost_pred/(p_9/r_9))%>%
    select(!contains("0")&!contains("1")&!contains("9"))
  
  # STEP 8 Consolidate outputs into comprehensive dataframes
  # if first iteration, nothing to append to seed dataframe
  if (phi_counter==1){ 
    io_exposure<-exposures%>%mutate(scenario=phi_counter,phi=phi)
    io_coef<-glm_coef%>%mutate(scenario=phi_counter,phi=phi)
    io_cor<-cor_df%>%mutate(scenario=phi_counter,phi=phi)
    io_gini<-gini%>%mutate(scenario=phi_counter,phi=phi)
    io_ave<-ave%>%mutate(scenario=phi_counter,phi=phi)
    # if second or greater iteration, append to previous iteration
  } else {
    io_exposure<-io_exposure%>%bind_rows( exposures%>%mutate(scenario=phi_counter,phi=phi) )
    io_coef<-io_coef%>%bind_rows(glm_coef%>%mutate(scenario=phi_counter,phi=phi))
    io_cor<-io_cor%>%bind_rows(cor_df%>%mutate(scenario=phi_counter,phi=phi))
    io_gini<-io_gini%>%bind_rows(gini%>%mutate(scenario=phi_counter,phi=phi))
    io_ave<-io_ave%>%bind_rows(ave%>%mutate(scenario=phi_counter,phi=phi))
  }
  
  # output message to console with progress through loop
  print(paste('iteration',phi_counter,'complete'))
  rm(exposures,exposures_holdout,glm_coef,gini,cor_df,
     alpha,delta,epsilon,gamma_scale,gamma_shape,
     lambda_avg,n_max,pct_holdout,pct_protected,phi)
     
  gc()}


